from importlib.metadata import version

__version__ = version(__package__)

from .api import Strats as Strats
from .core import Clock as Clock
from .core import Data as Data
from .core import Monitor as Monitor
from .core import QueueMsg as QueueMsg
from .core import State as State
from .core import Strategy as Strategy



from urllib.parse import urljoin

import httpx
import typer

from .common import DEFAULT_URL

app = typer.Typer(no_args_is_help=True)


@app.command()
def show():
    url = urljoin(DEFAULT_URL, "/clock")
    res = httpx.get(url)
    show_status(res)


@app.command()
def start():
    url = urljoin(DEFAULT_URL, "/clock/start")
    res = httpx.post(url)
    show_status(res)


@app.command()
def stop():
    url = urljoin(DEFAULT_URL, "/clock/stop")
    res = httpx.post(url)
    show_status(res)


def show_status(res):
    if res.status_code != 200:
        typer.echo(f"API reqeust failed: {res.text}", err=True)
        raise typer.Exit()

    data = res.json()

    print(f"current: {data['datetime']}")
    print(f"is_real: {data['is_real']}")
    if not data["is_real"]:
        print(f"is_running: {data['is_running']}")



from .clock import Clock as Clock
from .data import Data as Data
from .data import QueueMsg as QueueMsg
from .monitor import Monitor as Monitor
from .state import State as State
from .strategy import Strategy as Strategy



from zoneinfo import ZoneInfo

from strats import Clock, Strats


def main():
    Strats(
        clock=Clock(
            start_at="2025-01-01 12:00:00",
            tz=ZoneInfo("Asia/Tokyo"),
            speed=2,
        ),
    ).serve()


if __name__ == "__main__":
    main()



import asyncio
import logging
from datetime import datetime
from typing import Callable, Optional

from croniter import croniter

from strats.core import Clock, Monitor, State

logger = logging.getLogger(__name__)


class CronMonitor(Monitor):
    def __init__(
        self,
        cron_job: Callable,
        cron_schedule: str,
        check_interval_sec: float = 0.5,
        **kwargs,
    ):
        self.cron_job = cron_job
        self.cron_schedule = cron_schedule
        self.check_interval_sec = check_interval_sec

        super().__init__(**kwargs)

    async def run(self, clock: Clock, state: Optional[State]):
        await self.delay()
        self.set_descriptor(state)
        logger.info(f"{self.name} start")

        success = self.exec_on_init()
        if not success:
            return

        schedule = croniter(self.cron_schedule, clock.datetime)
        try:
            next_time = schedule.get_next(datetime)
            while True:
                if clock.datetime >= next_time:
                    # Exec job
                    source = self.cron_job(clock, state)

                    # Update state and lifecycle hooks
                    self.exec_on_pre_event(source)
                    self.update_data_descriptor(state, source)
                    self.exec_on_post_event(source)

                    # prepare next schedule
                    next_time = schedule.get_next(datetime)
                await asyncio.sleep(self.check_interval_sec)

        except asyncio.CancelledError:
            # To avoid "ERROR:asyncio:Task exception was never retrieved",
            # Re-raise the CancelledError
            raise
        except Exception as e:
            # Unexpected error
            logger.error(f"Error in {self.name}, but maybe in the `cron_job` function: {e}")
        finally:
            self.exec_on_delete()
            logger.info(f"{self.name} stopped")



import logging

from fastapi import APIRouter, Depends, HTTPException, Response
from prometheus_client import CONTENT_TYPE_LATEST, generate_latest

from strats.core.kernel import Kernel

logger = logging.getLogger(__name__)

router = APIRouter()


def get_kernel() -> Kernel:
    raise NotImplementedError("get_kernel is not yet bound")


@router.get("/healthz")
def healthz():
    return "ok"


@router.get("/metrics")
def metrics():
    data = generate_latest()
    return Response(content=data, media_type=CONTENT_TYPE_LATEST)


@router.get("/strategy", tags=["strategy"])
def get_strategy(kernel: Kernel = Depends(get_kernel)):
    return response_strategy_info(kernel)


@router.post("/strategy/start", tags=["strategy"])
async def start_strategy(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.start_strategy()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_strategy_info(kernel)


@router.post("/strategy/stop", tags=["strategy"])
async def stop_strategy(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.stop_strategy()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_strategy_info(kernel)


@router.get("/monitors", tags=["monitors"])
def get_monitors(kernel: Kernel = Depends(get_kernel)):
    return response_monitors_info(kernel)


@router.post("/monitors/start", tags=["monitors"])
async def start_monitors(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.start_monitors()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_monitors_info(kernel)


@router.post("/monitors/stop", tags=["monitors"])
async def stop_monitors(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.stop_monitors()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_monitors_info(kernel)


@router.get("/clock", tags=["clock"])
def get_clock(kernel: Kernel = Depends(get_kernel)):
    return response_clock_info(kernel)


@router.post("/clock/start", tags=["clock"])
async def start_clock(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.start_clock()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_clock_info(kernel)


@router.post("/clock/stop", tags=["clock"])
async def stop_clock(kernel: Kernel = Depends(get_kernel)):
    try:
        await kernel.stop_clock()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return response_clock_info(kernel)


def response_strategy_info(kernel):
    return {
        "is_configured": kernel.strategy is not None,
        "is_running": (kernel.strategy_task is not None and not kernel.strategy_task.done()),
    }


def response_monitors_info(kernel):
    res = {
        "is_configured": kernel.monitors is not None,
    }
    if kernel.monitors is not None:
        res["monitors"] = {
            monitor.name: {
                "is_running": (
                    monitor.name in kernel.monitor_tasks
                    and not kernel.monitor_tasks[monitor.name].done()
                )
            }
            for monitor in kernel.monitors
        }
    return res


def response_clock_info(kernel):
    if kernel.clock.is_mock:
        return {
            "is_real": False,
            "is_running": (kernel.clock_task is not None and not kernel.clock_task.done()),
            "datetime": kernel.clock.datetime.isoformat(),
        }
    else:
        return {
            "is_real": True,
            "datetime": kernel.clock.datetime.isoformat(),
        }



from collections.abc import AsyncGenerator

from strats import Strats
from strats.monitor import StreamClient, StreamMonitor


class SampleStreamClient(StreamClient):
    async def stream(self) -> AsyncGenerator[int]:
        raise ValueError("ERROR IN STREAM_CLIENT")
        yield 1


def main():
    Strats(
        monitors=[
            StreamMonitor(
                client=SampleStreamClient(),
            )
        ],
    ).serve()


if __name__ == "__main__":
    main()



import typer

from strats import __version__

from .clock import app as clock_app
from .monitors import app as monitors_app
from .strategy import app as strategy_app

app = typer.Typer(
    rich_markup_mode=None,
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)
app.add_typer(strategy_app, name="strategy")
app.add_typer(monitors_app, name="monitors")
app.add_typer(clock_app, name="clock")


@app.command()
def version():
    print(f"strats version: {__version__}")


if __name__ == "__main__":
    app()



import logging
from logging.config import dictConfig

import uvicorn
from fastapi import FastAPI

from strats.core.kernel import Kernel
from strats.internal.log_config import LOGGING_CONFIG

from .router import get_kernel, router

BANNER = r"""
 _______ _______  ______ _______ _______ _______
 |______    |    |_____/ |_____|    |    |______
 ______|    |    |     \ |     |    |    ______|
"""


def kernel_getter_factory(kernel):
    def kernel_getter():
        return kernel

    return kernel_getter


class Strats(Kernel):
    def serve(self, host="0.0.0.0", port=8000):
        dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger(__name__)

        app = FastAPI()
        app.include_router(router)
        app.dependency_overrides[get_kernel] = kernel_getter_factory(self)

        logger.info(BANNER)

        # Use lower-level uvicorn API to handle SIGINT/SIGTERM properly
        config = uvicorn.Config(
            app=app,
            host=host,
            port=port,
            log_config=LOGGING_CONFIG,
        )
        server = uvicorn.Server(config)
        try:
            server.run()
        except KeyboardInterrupt:
            pass



import asyncio
from collections.abc import AsyncGenerator
from decimal import Decimal

from strats import Data, State, Strats
from strats.model import (
    PricesData,
    PricesMetrics,
    prices_data_to_prices_metrics,
)
from strats.monitor import StreamClient, StreamMonitor


class SampleState(State):
    prices = Data(
        data=PricesData(),
        metrics=PricesMetrics(),
        data_to_metrics=prices_data_to_prices_metrics,
    )


class SampleStreamClient(StreamClient):
    async def stream(self) -> AsyncGenerator[PricesData]:
        for i in range(10):
            yield PricesData(
                bid=Decimal("100") + Decimal(i),
                ask=Decimal("101") + Decimal(i),
            )
            await asyncio.sleep(10)


def main():
    Strats(
        state=SampleState(),
        monitors=[
            StreamMonitor(
                data_name="prices",
                client=SampleStreamClient(),
            )
        ],
    ).serve()


if __name__ == "__main__":
    main()



import asyncio
import queue
import re
import threading
import time
from urllib.parse import urljoin

import pytest
import requests

BASE_URL = "http://localhost:8000"
APPLICATION_FILEPATH = "tests/e2e/e04_strategy_state_stream_monitor/app.py"


@pytest.mark.asyncio
async def test_app(app_process_factory):
    proc = app_process_factory(APPLICATION_FILEPATH)

    try:
        # >> healthz, metrics

        res = requests.get(urljoin(BASE_URL, "/healthz"))
        assert res.status_code == 200
        assert res.json() == "ok"

        res = requests.get(urljoin(BASE_URL, "/metrics"))
        assert res.status_code == 200

        # >> strategy

        res = requests.get(urljoin(BASE_URL, "/strategy"))
        expect = {"is_configured": True, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect

        # >> monitors

        res = requests.get(urljoin(BASE_URL, "/monitors"))
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        # >> run

        res = requests.post(urljoin(BASE_URL, "/strategy/start"))
        expect = {"is_configured": True, "is_running": True}
        assert res.status_code == 200
        assert res.json() == expect

        res = requests.post(urljoin(BASE_URL, "/monitors/start"))
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": True,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        await asyncio.sleep(0.5)

        # >> check

        res = requests.get(urljoin(BASE_URL, "/metrics"))
        assert res.status_code == 200
        assert extract_unlabeled_metric_value(res.text, "prices_bid") == 100.0
        assert extract_unlabeled_metric_value(res.text, "prices_ask") == 101.0
        assert extract_unlabeled_metric_value(res.text, "prices_spread") == 1.0
        assert extract_unlabeled_metric_value(res.text, "prices_update_count_total") == 1.0

        stderrs = get_stderr_list(proc)
        # the last stdout is "GET /metrics HTTP/1.1 200 OK"
        assert "INFO : __main__ : strategy > bid: 100" in stderrs[-2]

        # >> stop

        res = requests.post(urljoin(BASE_URL, "/monitors/stop"))
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        res = requests.post(urljoin(BASE_URL, "/strategy/stop"))
        expect = {"is_configured": True, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect

    finally:
        proc.terminate()
        proc.wait()


def extract_unlabeled_metric_value(body: str, metric_name: str) -> float:
    pattern = rf"^{re.escape(metric_name)} ([0-9.e+-]+)"
    match = re.search(pattern, body, re.MULTILINE)
    if not match:
        raise ValueError(f"Metric {metric_name} not found")
    return float(match.group(1))


def read_stderr_lines(process, output_queue):
    for line in process.stderr:
        output_queue.put(line)
    output_queue.put(None)  # signal end of stream


def get_stderr_list(process, timeout=1) -> list[str]:
    stderrs = []

    q: queue.Queue = queue.Queue()
    t = threading.Thread(target=read_stderr_lines, args=(process, q), daemon=True)
    t.start()

    start_time = time.time()
    while True:
        try:
            line = q.get(timeout=0.1)
            if line is None:  # end of stream
                break
            stderrs.append(line.strip())
        except queue.Empty:
            if time.time() - start_time > timeout:
                break
    return stderrs



import asyncio
import logging
from datetime import datetime, timedelta
from typing import Optional
from zoneinfo import ZoneInfo

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Clock:
    def __init__(
        self,
        *,
        tz: Optional[ZoneInfo] = None,
        start_at: Optional[str] = None,
        speed: float = 1.0,
    ):
        self.tz = tz
        self.speed = speed
        self.is_mock = start_at is not None

        if self.is_mock:
            t = datetime.strptime(str(start_at), DATETIME_FORMAT)
            if tz:
                t = t.replace(tzinfo=tz)
            self.mock_datetime = t

    @property
    def datetime(self):
        if self.is_mock:
            return self.mock_datetime

        if self.tz:
            return datetime.now(self.tz)
        else:
            return datetime.now()

    @property
    def ohlc_datetime(self):
        return self.datetime.replace(second=0, microsecond=0)

    async def run(self):
        if not self.is_mock:
            return

        logger.info("mock clock start")
        try:
            while True:
                await asyncio.sleep(1 / self.speed)
                self.mock_datetime += timedelta(seconds=1)
        except asyncio.CancelledError:
            pass
        finally:
            logger.info("mock clock stopped")



import subprocess
import sys
import time

import pytest


@pytest.fixture(scope="function")
def app_process_factory():
    # The argument is application filepath
    def _start(filepath: str):
        proc = subprocess.Popen(
            ["python", filepath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
        )
        time.sleep(0.5)

        if proc.poll() is not None:
            stdout, stderr = proc.communicate()
            print("[STDOUT]", stdout.encode(), file=sys.stderr)
            print("[STDERR]", stderr.encode(), file=sys.stderr)
            raise RuntimeError("Application process exited early")

        return proc

    yield _start



import asyncio
from urllib.parse import urljoin

import pytest
import requests

BASE_URL = "http://localhost:8000"
APPLICATION_FILEPATH = "tests/e2e/e02_cron_monitor/app.py"


@pytest.mark.asyncio
async def test_app(app_process_factory):
    proc = app_process_factory(APPLICATION_FILEPATH)

    try:
        # >> healthz, metrics

        res = requests.get(urljoin(BASE_URL, "/healthz"))
        assert res.status_code == 200
        assert res.json() == "ok"

        res = requests.get(urljoin(BASE_URL, "/metrics"))
        assert res.status_code == 200

        # >> strategy

        res = requests.get(urljoin(BASE_URL, "/strategy"))
        expect = {"is_configured": False, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect

        res = requests.post(urljoin(BASE_URL, "/strategy/start"))
        expect = {"detail": "Missing strategy configuration"}
        assert res.status_code == 400
        assert res.json() == expect

        res = requests.post(urljoin(BASE_URL, "/strategy/stop"))
        expect = {"detail": "Missing strategy configuration"}
        assert res.status_code == 400
        assert res.json() == expect

        # >> monitors

        res = requests.get(urljoin(BASE_URL, "/monitors"))
        expect = {
            "is_configured": True,
            "monitors": {
                "CronMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        res = requests.post(urljoin(BASE_URL, "/monitors/start"))
        expect = {
            "is_configured": True,
            "monitors": {
                "CronMonitor_1": {
                    "is_running": True,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        await asyncio.sleep(0.5)

        res = requests.post(urljoin(BASE_URL, "/monitors/stop"))
        expect = {
            "is_configured": True,
            "monitors": {
                "CronMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

    finally:
        proc.terminate()
        proc.wait()



from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

from prometheus_client import Counter, Gauge


@dataclass
class PricesData:
    bid: Decimal = Decimal("0")
    ask: Decimal = Decimal("0")


class PricesMetrics:
    def __init__(self, prefix: Optional[str] = None):
        p = "" if prefix is None else f"{prefix}_"
        self.bid = Gauge(p + "prices_bid", "")
        self.ask = Gauge(p + "prices_ask", "")
        self.spread = Gauge(p + "prices_spread", "")
        self.update_count = Counter(p + "prices_update_count", "")


def prices_data_to_prices_metrics(data: PricesData, metrics: PricesMetrics):
    metrics.bid.set(float(data.bid))
    metrics.ask.set(float(data.ask))
    metrics.spread.set(float(data.ask - data.bid))
    metrics.update_count.inc()



import asyncio
import logging
import queue
import threading

from strats.internal.lru_set import LRUSet

logger = logging.getLogger(__name__)


class State:
    def __init__(self, dedup_cache_size=100):
        self.lruset = LRUSet(capacity=dedup_cache_size)

    def set_queues(self):
        """
        set_queues initializes both synchronous and asynchronous queues.

        To avoid attaching them to the default event loop,
        this function must be called after the FastAPI server has started.

        The State class has embedded queue and sync_queue. These serve as pipes
        connecting Data, State, and external components. The reason why State
        (rather than Kernel) holds these queues is that Data can only access State.
        """
        if not hasattr(self, "_initialized"):
            self.sync_queue = queue.Queue()
            self.queue = asyncio.Queue()
            self._initialized = True

    def flush_queue(self):
        self.queue = asyncio.Queue()

    def run(self, stop_event: threading.Event):
        loop = asyncio.get_running_loop()
        self.sync_to_async_queue_thread = threading.Thread(
            target=self._sync_to_async_queue,
            args=(loop, stop_event),
        )
        self.sync_to_async_queue_thread.start()

    def _sync_to_async_queue(
        self,
        loop: asyncio.AbstractEventLoop,
        stop_event: threading.Event,
    ):
        logger.info("sync_to_async_queue thread start")

        while not stop_event.is_set():
            try:
                item = self.sync_queue.get(timeout=1)
            except queue.Empty:
                continue

            if item is None:
                break  # the stop signal

            # dedug filter
            if self.lruset.contains(item):
                continue
            self.lruset.add(item)

            # When scheduling callbacks from another thread,
            # `call_soon_threadsafe` must be used, since `call_soon` is not thread-safe.
            # cf. https://docs.python.org/ja/3.13/library/asyncio-eventloop.html#asyncio.loop.call_soon_threadsafe
            loop.call_soon_threadsafe(self.queue.put_nowait, item)

        logger.info("sync_to_async_queue thread stopped")



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Callable, Optional

from .clock import Clock
from .state import State

logger = logging.getLogger(__name__)


class Monitor(ABC):
    """Base class for monitoring functionality"""

    _counter = 0

    def __init__(
        self,
        name: Optional[str] = None,
        data_name: Optional[str] = None,
        start_delay_seconds: int = 0,
        # Lifecycle Hook
        on_init: Optional[Callable] = None,
        on_delete: Optional[Callable] = None,
        on_pre_event: Optional[Callable] = None,
        on_post_event: Optional[Callable] = None,
    ):
        # Update class-specific counter
        type(self)._counter += 1

        # Initialize common attributes
        self.name = name or f"{type(self).__name__}_{type(self)._counter}"
        self.data_name = data_name
        self.start_delay_seconds = start_delay_seconds

        # Set up lifecycle hooks
        self.on_init = on_init
        self.on_delete = on_delete
        self.on_pre_event = on_pre_event
        self.on_post_event = on_post_event

    @abstractmethod
    async def run(self, clock: Clock, state: Optional[State]):
        pass

    async def delay(self):
        if self.start_delay_seconds > 0:
            await asyncio.sleep(self.start_delay_seconds)

    def set_descriptor(self, state: Optional[State]):
        self.data_descriptor = None
        if state is not None and self.data_name:
            if self.data_name in type(state).__dict__:
                self.data_descriptor = type(state).__dict__[self.data_name]
            else:
                raise ValueError(f"data_name: `{self.data_name}` is not found in State")

    def exec_hook(self, f: Optional[Callable], f_name: str, *args) -> bool:
        if f is None:
            return True
        try:
            f(*args)
            return True
        except Exception as e:
            logger.error(f"lifecycle `{f_name}` error in {self.name}: {e}")
            return False

    def exec_on_init(self) -> bool:
        return self.exec_hook(self.on_init, "on_init")

    def exec_on_delete(self) -> bool:
        return self.exec_hook(self.on_delete, "on_delete")

    def exec_on_pre_event(self, source) -> bool:
        return self.exec_hook(self.on_pre_event, "on_pre_event", source)

    def exec_on_post_event(self, source) -> bool:
        return self.exec_hook(self.on_post_event, "on_post_event", source)

    def update_data_descriptor(self, state, source):
        if self.data_descriptor is None:
            return True
        try:
            self.data_descriptor.__set__(state, source)
            return True
        except Exception as e:
            logger.error(f"failed to update state.{self.data_name} in {self.name}: {e}")
            return False
